use std::{
    fs::{self, File, OpenOptions},
    io::{ErrorKind, Read as _, Result, Write as _},
    path::Path,
};

use crate::error::{NetavarkError, NetavarkResult};

pub fn apply_sysctl_value(ns_value: impl AsRef<str>, val: impl AsRef<str>) -> NetavarkResult<()> {
    _apply_sysctl_value(&ns_value, val, &mut None::<SysctlDWriter<&str>>)
        .map_err(|e| NetavarkError::wrap(format!("set sysctl {}", ns_value.as_ref()), e.into()))
}

pub fn apply_sysctl_value_with_writer(
    ns_value: impl AsRef<str>,
    val: impl AsRef<str>,
    writer: &mut Option<SysctlDWriter<impl AsRef<Path>>>,
) -> NetavarkResult<()> {
    _apply_sysctl_value(&ns_value, val, writer)
        .map_err(|e| NetavarkError::wrap(format!("set sysctl {}", ns_value.as_ref()), e.into()))
}

/// Set a sysctl value by value's namespace.
/// ns_value is the path of the sysctl (using slashes not dots!) and without the "/proc/sys/" prefix.
fn _apply_sysctl_value(
    ns_value: impl AsRef<str>,
    val: impl AsRef<str>,
    sysctld_writer: &mut Option<SysctlDWriter<impl AsRef<Path>>>,
) -> Result<()> {
    const PREFIX: &str = "/proc/sys/";
    let ns_value = ns_value.as_ref();
    let mut path = String::with_capacity(PREFIX.len() + ns_value.len());
    path.push_str(PREFIX);
    path.push_str(ns_value);
    let val = val.as_ref();

    log::debug!("Setting sysctl value for {} to {}", ns_value, val);

    if let Some(writer) = sysctld_writer {
        writer.write(ns_value, val)?;
    }

    let mut f = File::open(&path)?;
    let mut buf = String::with_capacity(1);
    f.read_to_string(&mut buf)?;

    if buf.trim() == val {
        return Ok(());
    }

    let mut f = OpenOptions::new().write(true).open(&path)?;
    f.write_all(val.as_bytes())
}

pub fn disable_ipv6_autoconf(if_name: &str) -> NetavarkResult<()> {
    // make sure autoconf is off, we want manual config only
    if let Err(err) = _apply_sysctl_value(
        format!("net/ipv6/conf/{if_name}/autoconf"),
        "0",
        &mut None::<SysctlDWriter<&str>>,
    ) {
        match err.kind() {
            ErrorKind::NotFound => {
                // if the sysctl is not found we likely run on a system without ipv6
                // just ignore that case
            }

            // if we have a read only /proc we ignore it as well
            ErrorKind::ReadOnlyFilesystem => {}

            _ => {
                return Err(NetavarkError::wrap(
                    "failed to set autoconf sysctl",
                    err.into(),
                ));
            }
        }
    };
    Ok(())
}

pub fn get_bridge_sysctl_d_path(bridge_name: &str) -> String {
    format!("/run/sysctl.d/10-netavark-{bridge_name}.conf")
}

pub struct SysctlDWriter<P: AsRef<Path>> {
    f: File,
    path: P,
    commit: bool,
}

impl<P: AsRef<Path>> SysctlDWriter<P> {
    /// create a new sysctl writer for the given file, returns None if unable to write the file.
    /// Important one must call commit() otherwise we remove the file again on drop.
    pub fn new(path: P) -> Option<Self> {
        // Note using create_new here as we don't want to overwrite an existing file
        // If such file already exists it should be safe to assume it has the right
        // content so we can skip writing it again.
        let mut f = File::create_new(&path).ok()?;
        f.write_all("# autogenerated by netavark\n".as_bytes())
            .ok()?;
        Some(Self {
            f,
            path,
            commit: false,
        })
    }

    /// write the sysctl in format according to sysctl.d(5)
    /// Important one must call commit() otherwise we remove the file again on drop.
    fn write(&mut self, ns_value: &str, val: &str) -> Result<()> {
        let mut entry = String::with_capacity(ns_value.len() + val.len() + 4);
        entry.push_str(ns_value);
        entry.push_str(" = ");
        entry.push_str(val);
        entry.push('\n');
        self.f.write_all(entry.as_bytes())?;
        Ok(())
    }

    /// mark as successfully written so the file is not removed on drop
    /// The reason we do this is to avoid manually removing the file on
    /// each early exit due errors.
    /// This also consumes the writer so it is impossible to keep writing
    /// after this is called.
    pub fn commit(mut self) {
        self.commit = true;
    }
}

impl<P: AsRef<Path>> Drop for SysctlDWriter<P> {
    fn drop(&mut self) {
        if self.commit {
            // do nothing
            return;
        }
        // ignore errors here
        let _ = fs::remove_file(&self.path);
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use tempfile::Builder;

    #[test]
    fn test_sysctl_d_writer() {
        let tmp_dir = Builder::new()
            .prefix("example")
            .tempdir()
            .expect("failed to create tmpdir");
        let filepath = tmp_dir.path().join("test.conf");
        let mut w = SysctlDWriter::new(&filepath).expect("failed to get writer");
        w.write("abc", "1").expect("failed to write value 1");
        w.write("def", "2").expect("failed to write value 2");
        w.commit();
        let res = fs::read_to_string(filepath).expect("failed to read file");
        assert_eq!(res, "# autogenerated by netavark\nabc = 1\ndef = 2\n");
    }

    #[test]
    fn test_sysctl_d_writer_nocommit() {
        let tmp_dir = Builder::new()
            .prefix("example")
            .tempdir()
            .expect("failed to create tmpdir");
        let filepath = tmp_dir.path().join("test.conf");
        let mut w = SysctlDWriter::new(&filepath).expect("failed to get writer");
        w.write("abc", "1").expect("failed to write value 1");
        // drop the writer here which should delete the file
        drop(w);
        assert!(
            !filepath.exists(),
            "file should not exist after drop without commit()"
        )
    }
}
